/**
 * Created by IntelliJ IDEA.
 * User: chizhik
 * Date: 08.11.2002
 * Time: 11:58:21
 */
package ru.ezhiki.automat.human.logic;

import org.xml.sax.*;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;

import javax.vecmath.Vector3d;
import java.io.FileReader;
import java.util.*;


public class FigureLoader extends DefaultHandler
{
    private HashMap joints = new HashMap();
    private Vector jointsVector = new Vector();

    private Joint centralJoint = null;

    private String currentJointName = null;
    private Joint currentJoint = null;

    private Figure figure = null;

    public void startElement(String namespaceURI,
                             String localName,
                             String qName,
                             Attributes attr) throws SAXException
    {
        if (localName.equals("figure"))
        {
            figure = new Figure(attr.getValue("name"), attr.getValue("authors"), attr.getValue("date"));

            joints.clear();
        }
        else if (localName.equals("joint"))
        {
            currentJointName = new String(attr.getValue("name"));
            currentJoint = new Joint();

            currentJoint.name = currentJointName;

            if (!attr.getValue("parent").equals("null"))
                currentJoint.parent = (Joint) joints.get(attr.getValue("parent"));

            currentJoint.m = new Double(attr.getValue("mass")).doubleValue();
            currentJoint.length = new Double(attr.getValue("length")).doubleValue();
        }
        else if (localName.equals("location"))
        {
            currentJoint.loc.x = new Double(attr.getValue("x")).doubleValue();
            currentJoint.loc.y = new Double(attr.getValue("y")).doubleValue();
            currentJoint.loc.z = new Double(attr.getValue("z")).doubleValue();
        }
        else if (localName.equals("massCentre"))
        {
            currentJoint.mLoc.x = new Double(attr.getValue("x")).doubleValue();
            currentJoint.mLoc.y = new Double(attr.getValue("y")).doubleValue();
            currentJoint.mLoc.z = new Double(attr.getValue("z")).doubleValue();
        }
        else if (localName.equals("angles"))
        {
            currentJoint.angles.x = new Double(attr.getValue("x")).doubleValue()/180*Math.PI;
            currentJoint.angles.y = new Double(attr.getValue("y")).doubleValue()/180*Math.PI;
            currentJoint.angles.z = new Double(attr.getValue("z")).doubleValue()/180*Math.PI;
        }
        else if (localName.equals("minAngles"))
        {
            currentJoint.minAngles.x = new Double(attr.getValue("x")).doubleValue()/180*Math.PI;
            currentJoint.minAngles.y = new Double(attr.getValue("y")).doubleValue()/180*Math.PI;
            currentJoint.minAngles.z = new Double(attr.getValue("z")).doubleValue()/180*Math.PI;
        }
        else if (localName.equals("maxAngles"))
        {
            currentJoint.maxAngles.x = new Double(attr.getValue("x")).doubleValue()/180*Math.PI;
            currentJoint.maxAngles.y = new Double(attr.getValue("y")).doubleValue()/180*Math.PI;
            currentJoint.maxAngles.z = new Double(attr.getValue("z")).doubleValue()/180*Math.PI;
        }
    }

    public void endElement(String namespaceURI,
                           String localName,
                           String qName) throws SAXException
    {
        if (localName.equals("figure"))
        {
            System.gc();

            figure.setJoints(jointsVector);

            figure.centralJoint = centralJoint;
        }
        else if (localName.equals("joint"))
        {
            Joint added = null;

            if (currentJoint.parent != null)
            {
                added = currentJoint.parent.addChild(currentJoint);
                added.name = currentJointName;
                joints.put(currentJointName, added);
                jointsVector.add(added);
            }
            else
            {
                added = new Joint(null,
                        (Vector3d) currentJoint.loc.clone(),
                        currentJoint.length,
                        currentJoint.angles,
                        currentJoint.minAngles,
                        currentJoint.maxAngles,
                        currentJoint.m,
                        currentJoint.mLoc);
                added.name = currentJointName;
                joints.put(currentJointName, added);
                jointsVector.add(added);
            }

            if (currentJointName.equals("centralJoint"))
                centralJoint = added;
        }
    }

    public FigureLoader()
    {
    }

    public Figure loadFile(String fName)
    {
        try
        {
            XMLReader reader = XMLReaderFactory.createXMLReader("org.apache.crimson.parser.XMLReaderImpl");
            reader.setContentHandler(this);
            reader.parse(new InputSource(new FileReader(fName)));
        }
        catch (SAXParseException e)
        {
            System.out.println("Error at line: " + e.getLineNumber() + ", column: " + e.getColumnNumber() + ", Id: " + e.getPublicId());
            e.printStackTrace();
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }

        return figure;
    }
}
