/*
 * Created by IntelliJ IDEA.
 * User: Alexander
 * Date: 20.10.2002
 * Time: 11:49:37
 */
package ru.ezhiki.automat.human.logic;

import ru.ezhiki.utils.MathUtils;
import ru.ezhiki.utils.Signaller;

import javax.vecmath.*;
import javax.media.j3d.*;
import java.util.Vector;
import java.util.Date;
import java.util.logging.Logger;

/**
 * The main class. It holds a joint representation. Joint is a stick, that is connected to other joints.
 * It turns, moves with the help of a simple automat. Joint is a part of Figure.
 * @see Figure
 * @see Actions
 */
public class Joint extends Thread
{
    /**
     * Name of the joint (To search it).
     */
    public String name = "";

    /**
     * Parent of this Joint. Central joint has no parent.
     */
    public Joint parent = null;

    /**
     * Children.
     */
    public Vector children = null;

    /**
     * length of the joint.
     */
    public double length = 0;

    /**
     * Minimum angles of the joint.
     */
    public Vector3d minAngles = new Vector3d(0, 0, 0);
    /**
     * Maximum angles of the joint.
     */
    public Vector3d maxAngles = new Vector3d(0, 0, 0);

    /**
     * Direction of the joint. X, Y - normal angles, Z - turn angle. These are relative, local angles.
     */
    public Vector3d angles = new Vector3d(0, 0, 0);

    /**
     * Position of the joint.
     */
    public Vector3d loc = new Vector3d(0, 0, 0);

    /**
     * Current settings of the angle.
     */
    private Vector3d curAngles = new Vector3d(0, 0, 0);
    /**
     * Current settings of the position.
     */
    private Vector3d curLoc = new Vector3d(0, 0, 0);

    /**
     * Mass centre of the joint.
     */
    public double m = 0;

    /**
     * Coordinates of the joint's mass centre
     */
    public Vector3d mLoc = null;

    /**
     * Java3D Group. Holds node and transform.
     */
    private BranchGroup branch = null;
    /**
     * Holds local transform of the joint.
     */
    private TransformGroup transform = null;

    /**
     * Delay between iterations while moving.
     */
    private static long delay = 4;

    /**
     * A variable for working (moving) signalling. Each dot is a working joint. If there are no dots - there are no working joints
     */
    private Signaller signaller = Signaller.getSignaller();

    /**
     * Speed of turning. Can be changed during moving. In radians per second.
     */
    private double speedTurn = MathUtils.deg2rad(15);

    /**
     * Speed of translating. Can be changed during moving. In location units per second.
     */
    private double speedTranslate = 2;

    /**
     * Timer for corrent moving speed calculation.
     */
    private double time = -1;

    /**
     * Viscosity of connection to parent. More viscosity means harder joint rotation about parent. Must be >= 0, <= 1.
     * 1 means, that this joint is connected to the parent so, that it can't rotate relatively to it. Otherwise, 0 means, that
     * it can rotate without any force from the parent.
     */
    private double viscosity = 0;

    /**
     * Inertia of the joint's end (???). How it reacts to the parent's moving. 0 means, that it reacts momentically, 1 - it doesn't
     * react (I mean, that the connection doesn't brake, but the other end of this joint does'n move)
     */
    private double inertia = 0;

    /**
     * Automat A0 state
     */
    private int y0 = 0;

    private Vector3d dA = new Vector3d();
    private Vector3d dL = new Vector3d();
    private Vector3d tA = new Vector3d();
    private Vector3d tL = new Vector3d();
    private Logger logger = null;

    /**
     * Constructor
     */
    public Joint(
            Joint parent,
            Vector3d loc,
            double length,
            Vector3d angles,
            Vector3d minAngles,
            Vector3d maxAngles,
            double m,
            Vector3d mLoc)
    {
        this.parent = parent;
        this.length = length;
        this.minAngles = minAngles;
        this.maxAngles = maxAngles;
        this.angles = angles;
        this.curAngles = (Vector3d)this.angles.clone();
        this.loc = loc;
        this.curLoc = (Vector3d)this.loc.clone();
        this.m = m;
        this.mLoc = mLoc;

        children = new Vector();

        if (this.mLoc == null)
        {
            this.mLoc = new Vector3d(0, 0, length/2);
            this.mLoc.add(curLoc);
        }

        setPriority(Thread.MIN_PRIORITY);

        start();
    }

    /**
     * Default constructor
     */
    public Joint()
    {
        this.parent = null;
        this.length = 5;
        this.m = 0;
        this.mLoc = null;
        children = new Vector();
    }

    /**
     * Main function of the thread
     */
    public void run()
    {
        while (true)
        {
            try
            {
                yield();
                sleep(delay);
            }
            catch (InterruptedException e)
            {
                e.printStackTrace();
            }

            if (x0_0() == true)
                A0();
        }
    }

    /**
     * Joint automat implementation.
     */
    void A0()
    {
        if (y0 == 0 && logger != null)
            logger.info(">>> A0 ( \"" + name + "\")");

        switch (y0)
        {
            case 0:
                if (x0_0() == true) { z0_0();               y0 = 1; }
            break;
            case 1:
                z0_1();
                if (x0_1() == true) { z0_2();               y0 = 0; }
            break;
        }

        if (y0 == 0 && logger != null)
            logger.info("<<< A0 ( \"" + name + "\")");
    }

    private boolean x0_0()
    {
        return (curLoc.x != loc.x || curLoc.y != loc.y || curLoc.z != loc.z ||
                            curAngles.x != angles.x || curAngles.y != angles.y || curAngles.z != angles.z);
    }

    private boolean x0_1()
    {
        return finishedUpdating(dA, tA, dL, tL);
    }

    private void z0_0()
    {
        if (logger != null)
            logger.info("    A0->z2_1 (   \"" + name +  "\" ...)");
        time = new Date().getTime();
        z0_1();
    }

    private void z0_1()
    {
        if (logger != null)
            logger.info(" %  A0->z2_1 (  \"" + name +  "\" ...)");
        smallMove();
        update();
    }

    private void z0_2()
    {
        if (logger != null)
            logger.info("    A0->z2_1 (   \"" + name +  "\")");
        signaller.removeSignal();
    }

    /**
     * Iterates angles during moving. And checks, did we reach the needed angles and location.
     */
    private void smallMove()
    {
        long curTime = new Date().getTime();
        double difference = (curTime - time)/1000.0;
        time = curTime;

        if (difference == 0)
            return;

        dA = (Vector3d)angles.clone();
        dA.sub(curAngles);
        tA = new Vector3d(speedTurn, speedTurn, speedTurn);

        dL = (Vector3d)loc.clone();
        dL.sub(curLoc);
        tL = new Vector3d(speedTranslate, speedTranslate, speedTranslate);

        if (dA.x != 0 || dA.y != 0 || dA.z != 0)
        {
            if (dA.x < 0)
                tA.x *= -1;
            if (dA.y < 0)
                tA.y *= -1;
            if (dA.z < 0)
                tA.z *= -1;

            tA.scale(difference);

            curAngles.x += updatedParameter(dA.x, tA.x);
            curAngles.y += updatedParameter(dA.y, tA.y);
            curAngles.z += updatedParameter(dA.z, tA.z);

            Vector3d checkedAngles = new Vector3d();

            checkedAngles.x = checkBounds(curAngles.x, maxAngles.x, minAngles.x);
            checkedAngles.y = checkBounds(curAngles.y, maxAngles.y, minAngles.y);
            checkedAngles.z = checkBounds(curAngles.z, maxAngles.z, minAngles.z);

            if (checkedAngles.x != curAngles.x)
            {
                dA.x = 0;
                curAngles.x = checkedAngles.x;
            }
            if (checkedAngles.y != curAngles.y)
            {
                dA.y = 0;
                curAngles.y = checkedAngles.y;
            }
            if (checkedAngles.z != curAngles.z)
            {
                dA.z = 0;
                curAngles.z = checkedAngles.z;
            }
        }

        if (dL.x != 0 || dL.y != 0 || dL.z != 0)
        {
            if (dL.x < 0)
                tL.x *= -1;
            if (dL.y < 0)
                tL.y *= -1;
            if (dL.z < 0)
                tL.z *= -1;

            tL.scale(difference);

            curLoc.x += updatedParameter(dL.x, tL.x);
            curLoc.y += updatedParameter(dL.y, tL.y);
            curLoc.z += updatedParameter(dL.z, tL.z);
        }
    }

    /**
     * Checks, whether first parameter is beetween second and third
     * @param x test parameter
     * @param maxX upper bound
     * @param minX lower bound
     * @return if minX <= x <= maxX, then it returns x, either - one of bounds
     */
    private double checkBounds(double x, double maxX, double minX)
    {
        if (x > maxX)
            x = maxX;
        else if (x < minX)
            x = minX;

        return x;
    }

    /**
     * Indicates, that angles and location do not need update more.
     */
    private boolean finishedUpdating(Vector3d dA, Vector3d tA, Vector3d dL, Vector3d tL)
    {
        if (Math.abs(dA.x) > Math.abs(tA.x) || Math.abs(dA.y) > Math.abs(tA.y) || Math.abs(dA.z) > Math.abs(tA.z) ||
                Math.abs(dL.x) > Math.abs(tL.x) || Math.abs(dL.y) > Math.abs(tL.y) || Math.abs(dL.z) > Math.abs(tL.z))
            return false;
        else
        {
            curLoc.set(loc);
            curAngles.set(angles);
            return true;
        }
    }

    /**
     * Choose, which absolute value is less. And returns it.
     */
    private double updatedParameter(double dX, double tX)
    {
        if (Math.abs(dX) > Math.abs(tX))
            return tX;
        else
            return dX;
    }

    /**
     * Updates transform (for visualizing), children and parent.
     */
    private void update()
    {
        transform.setTransform(getCurrentTransform());
        if (parent != null)
            parent.update();
    }

    /**
     * Returns Node for attaching into Java3D Scene.
     */
    public Node get3dShape()
    {
        branch = new BranchGroup();

        transform = new TransformGroup(getCurrentTransform());
        transform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);

        branch.addChild(transform);

        Material material = new Material();
        material.setLightingEnable(true);
        material.setDiffuseColor(0.8f, 0.8f, 0.8f);

        Appearance appearanceLines = new Appearance();
        appearanceLines.setColoringAttributes(new ColoringAttributes(new Color3f(0.5f, 0.5f, 0.5f), ColoringAttributes.SHADE_GOURAUD));
        appearanceLines.setLineAttributes(new LineAttributes(2, LineAttributes.PATTERN_SOLID, true));
        appearanceLines.setMaterial(material);

        Appearance appearancePoints = new Appearance();
        appearancePoints.setColoringAttributes(new ColoringAttributes(new Color3f(0.5f, 0.5f, 0.5f), ColoringAttributes.SHADE_GOURAUD));
        appearancePoints.setPointAttributes(new PointAttributes((int)(m/2 > 4 ? m/2 : 4), true));
        appearancePoints.setMaterial(material);

        IndexedPointArray points = new IndexedPointArray(1, GeometryArray.COORDINATES | GeometryArray.COLOR_3, 1);

        points.setColorIndices(0, new int[] {0});
        points.setColors(0, new Color3f[] {new Color3f(0.1f, 0.1f, 1)});
        points.setCoordinateIndices(0, new int[] {0});
        points.setCoordinates(0, new Point3d[] {new Point3d(mLoc)});

        IndexedLineArray lines =
                new IndexedLineArray(2, GeometryArray.COORDINATES | GeometryArray.COLOR_3, 2);

        Vector3d drawLoc = (Vector3d)curLoc.clone();
        Vector3d drawEndLoc = (Vector3d)curLoc.clone();
        drawEndLoc.add(new Vector3d(0, 0, length));

        int pntsIndex[] = {0, 1};
        Point3d pnts[] = {new Point3d(drawLoc), new Point3d(drawEndLoc)};
        int clrsIndex[] = {0, 1};
        Color3f clrs[] = {new Color3f(1, 0.1f, 0.1f), new Color3f(0.1f, 1, 0.1f)};

        lines.setColorIndices(0, clrsIndex);
        lines.setColors(0, clrs);
        lines.setCoordinateIndices(0, pntsIndex);
        lines.setCoordinates(0, pnts);

        transform.addChild(new Shape3D(lines, appearanceLines));
        transform.addChild(new Shape3D(points, appearancePoints));

        for (int i = 0; i < children.size(); i++)
        {
            Joint joint = (Joint) children.elementAt(i);
            transform.addChild(joint.get3dShape());
        }

        return branch;
    }

    /**
     * Current transform of this joint.
     * @return transform matrix
     */
    private Transform3D getCurrentTransform()
    {
        Transform3D toView = new Transform3D();
        Quat4d q = quatFromEuler();

        Vector3d coordStart = null;
        if (parent != null)
        {
            coordStart = (Vector3d)parent.curLoc.clone();
            coordStart.add(new Vector3d(0, 0, parent.length));
            coordStart.add(curLoc);
        }
        else
        {
            coordStart = (Vector3d)curLoc.clone();
        }

        toView.setRotation(q);
        toView.setTranslation(coordStart);

        return toView;
    }

    /**
     * Returns quaternion, that is constructed from angles field.
     */
    private Quat4d quatFromEuler()
    {
        double roll = curAngles.x;
        double pitch = curAngles.y;
        double yaw = curAngles.z;

        Quat4d qRoll = new Quat4d(Math.sin(roll/2), 0, 0, Math.cos(roll/2));
        Quat4d qPitch = new Quat4d(0, Math.sin(pitch/2), 0, Math.cos(pitch/2));
        Quat4d qYaw = new Quat4d(0, 0, Math.sin(yaw/2), Math.cos(yaw/2));

        Quat4d q = new Quat4d();

        q = qYaw;
        q.mul(qPitch);
        q.mul(qRoll);

        return q;
    }

    /**
     * Adds a child to this joint.
     * @return a joint, that was added. Notice, that it is not the same joint, that was passed as a parameter.
     */
    public Joint addChild(Joint aJoint)
    {
        Joint automat = new Joint(this, aJoint.loc, aJoint.length, aJoint.angles, aJoint.minAngles, aJoint.maxAngles, aJoint.m, aJoint.mLoc);
        children.add(automat);

        return automat;
    }

    /**
     * Set x-angle
     */
    public void setAngleX(double _x)
    {
        if (angles.x != _x)
            angles.x = _x;
        else
            signaller.removeSignal();
    }

    /**
     * Set y-angle
     */
    public void setAngleY(double _y)
    {
        if (angles.y != _y)
            angles.y = _y;
        else
            signaller.removeSignal();
    }

    /**
     * Set z-angle
     */
    public void setAngleZ(double _z)
    {
        if (angles.z != _z)
            angles.z = _z;
        else
            signaller.removeSignal();
    }

    /**
     * Set x coordinate of the location
     */
    public void setLocX(double _x)
    {
        if (loc.x != _x)
            loc.x = _x;
        else
            signaller.removeSignal();
    }

    /**
     * Set y coordinate of the location
     */
    public void setLocY(double _y)
    {
        if (loc.y != _y)
            loc.y = _y;
        else
            signaller.removeSignal();
    }

    /**
     * Set z coordinate of the location
     */
    public void setLocZ(double _z)
    {
        if (loc.z != _z)
            loc.z = _z;
        else
            signaller.removeSignal();
    }

    /**
     * Set all angles and location.
     */
    public void setAnglesAndLocation(Vector3d a, Vector3d l)
    {
        boolean wasSet = false;

        if (a != null && (angles.x != a.x || angles.y != a.y || angles.z != a.z))
        {
            angles = (Vector3d)a.clone();
            wasSet = true;
        }
        if (l != null && (loc.x != l.x || loc.y != l.y || loc.z != l.z))
        {
            loc = (Vector3d)l.clone();
            wasSet = true;
        }

        if (!wasSet)
            signaller.removeSignal();
    }

    /**
     * Set speeds of turning and translation.
     */
    public void setSpeeds(String speedTurn, String speedTranslate)
    {
        setSpeedTurn(speedTurn);
        setSpeedTranslate(speedTranslate);
    }

    /**
     * Set speed of translation.
     */
    public void setSpeedTranslate(String speedTranslate)
    {
        if (speedTranslate != null)
        {
            try
            {
                this.speedTranslate = new Double(speedTranslate).doubleValue();
            }
            catch (NumberFormatException e)
            {
                this.speedTranslate = 2;
            }
        }
        else
        {
            this.speedTranslate = 2;
        }
    }

    /**
     * Set speed of turning.
     */
    public void setSpeedTurn(String speedTurn)
    {
        if (speedTurn != null)
        {
            try
            {
                this.speedTurn = MathUtils.deg2rad(new Double(speedTurn).doubleValue());
            }
            catch (NumberFormatException e)
            {
                this.speedTurn = MathUtils.deg2rad(15);
            }
        }
        else
        {
            this.speedTurn = MathUtils.deg2rad(15);
        }
    }

    public String getJointName()
    {
        return name;
    }

    public void setLogger(Logger logger)
    {
        this.logger = logger;
    }
}
